@@ -9,7 +9,9 @@
 #import "iTermApplicationDelegate.h"
 
 #import "iTermAdvancedSettingsModel.h"
+#import "iTermFakeUserDefaults.h"
 #import "iTermPasteHelper.h"
+#import "iTermSelectorSwizzler.h"
 #import "iTermWarning.h"
 #import "NSData+iTerm.h"
 #import "NSStringITerm.h"
@@ -126,7 +128,7 @@ static const double kFloatingPointTolerance = 0.00001;
                                                    regex:regex
                                             substitution:substitution];
     [iTermPasteHelper sanitizePasteEvent:event encoding:NSUTF8StringEncoding];
-    XCTAssert([expected isEqualToString:event.string]);
+    XCTAssertEqualObjects(expected, event.string);
 }
 
 - (void)testSanitizeIdentity {
@@ -221,7 +223,7 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:kHelloWorld
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     [self runTimer];
@@ -234,7 +236,7 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:kTestString
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     [self runTimer];
@@ -246,19 +248,19 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:@"\t"
                   slowly:NO
         escapeShellChars:YES
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformConvertToSpaces
             spacesPerTab:4];
     [self runTimer];
     NSString *expected = @"\\ \\ \\ \\ ";
-    XCTAssert([_writeBuffer isEqualToString:expected]);
+    XCTAssertEqualObjects(_writeBuffer, expected);
 }
 
 - (void)testEscapeDoesNotEscapeCarriageReturn {
     [_helper pasteString:@"\r"
                   slowly:NO
         escapeShellChars:YES
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformConvertToSpaces
             spacesPerTab:4];
     [self runTimer];
@@ -270,19 +272,19 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:kTestString
                   slowly:NO
         escapeShellChars:YES
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformConvertToSpaces
             spacesPerTab:4];
     [self runTimer];
     NSString *expected = @"a\\ \\(\\ \\ \\ \\ \r\r“”‘’–—b";
-    XCTAssert([_writeBuffer isEqualToString:expected]);
+    XCTAssertEqualObjects(_writeBuffer, expected);
 }
 
 - (void)testDoNotEscapeNonAscii {
     [_helper pasteString:@"“"
                   slowly:NO
         escapeShellChars:YES
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformEscapeWithCtrlV
             spacesPerTab:0];
     [self runTimer];
@@ -294,7 +296,7 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:@"\x16"
                   slowly:NO
         escapeShellChars:YES
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformEscapeWithCtrlV
             spacesPerTab:0];
     [self runTimer];
@@ -306,7 +308,7 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:kTestString
                   slowly:NO
         escapeShellChars:YES
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformEscapeWithCtrlV
             spacesPerTab:0];
     [self runTimer];
@@ -314,7 +316,21 @@ static const double kFloatingPointTolerance = 0.00001;
     XCTAssert([_writeBuffer isEqualToString:expected]);
 }
 
-- (void)testMultilineWarning {
+- (void)testMultilineWarningWithOverride {
+    [iTermSelectorSwizzler swizzleSelector:@selector(promptForPasteWhenNotAtPrompt)
+                                 fromClass:[iTermAdvancedSettingsModel class]
+                                 withBlock:^ BOOL { return YES; }
+                                  forBlock:^{
+                                      [iTermAdvancedSettingsModel loadAdvancedSettingsFromUserDefaults];
+                                      [self doMultilineWarningTestWithOverride:YES];
+                                  }];
+}
+
+- (void)testMultilineWarningNoOverride {
+    [self doMultilineWarningTestWithOverride:NO];
+}
+
+- (void)testSingleLinePasteGivesNoWarning {
     __block BOOL warned = NO;
     [_warningBlock release];
     _warningBlock = [^NSModalResponse(NSAlert *alert, NSString *identifier) {
@@ -323,44 +339,67 @@ static const double kFloatingPointTolerance = 0.00001;
         return 1;  /* deprecated NSAlertDefaultReturn; */
     } copy];
 
-    // Check cr newline
-    [_helper pasteString:@"line 1\rline 2"
-                  slowly:NO
-        escapeShellChars:NO
-                commands:NO
-            tabTransform:kTabTransformNone
-            spacesPerTab:0];
-    XCTAssert(!warned);
-
-    // Check lf newline
-    warned = NO;
-    [_helper pasteString:@"line 1\nline 2"
-                  slowly:NO
-        escapeShellChars:NO
-                commands:NO
-            tabTransform:kTabTransformNone
-            spacesPerTab:0];
-    XCTAssert(!warned);
-
-    // Check crlf newline
-    warned = NO;
-    [_helper pasteString:@"line 1\r\nline 2"
-                  slowly:NO
-        escapeShellChars:NO
-                commands:NO
-            tabTransform:kTabTransformNone
-            spacesPerTab:0];
-    XCTAssert(!warned);
-
     // Check no newline gives no warning.
     warned = NO;
     [_helper pasteString:@"line 1"
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     XCTAssert(!warned);
+
+}
+
+- (void)doMultilineWarningTestWithOverride:(BOOL)override {
+    __block BOOL warned = NO;
+    [_warningBlock release];
+    _warningBlock = [^NSModalResponse(NSAlert *alert, NSString *identifier) {
+        XCTAssert([identifier isEqualToString:[iTermAdvancedSettingsModel noSyncDoNotWarnBeforeMultilinePasteUserDefaultsKey]]);
+        warned = YES;
+        return 1;  /* deprecated NSAlertDefaultReturn; */
+    } copy];
+
+    // Check cr newline.
+    [_helper pasteString:@"line 1\rline 2"
+                  slowly:NO
+        escapeShellChars:NO
+                isUpload:NO
+            tabTransform:kTabTransformNone
+            spacesPerTab:0];
+    if (override) {
+        XCTAssert(warned);
+    } else {
+        XCTAssert(!warned);
+    }
+
+    // Check lf newline.
+    warned = NO;
+    [_helper pasteString:@"line 1\nline 2"
+                  slowly:NO
+        escapeShellChars:NO
+                isUpload:NO
+            tabTransform:kTabTransformNone
+            spacesPerTab:0];
+    if (override) {
+        XCTAssert(warned);
+    } else {
+        XCTAssert(!warned);
+    }
+
+    // Check crlf newline.
+    warned = NO;
+    [_helper pasteString:@"line 1\r\nline 2"
+                  slowly:NO
+        escapeShellChars:NO
+                isUpload:NO
+            tabTransform:kTabTransformNone
+            spacesPerTab:0];
+    if (override) {
+        XCTAssert(warned);
+    } else {
+        XCTAssert(!warned);
+    }
 }
 
 - (void)testBracketingOnPasteString {
@@ -368,7 +407,7 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:kHelloWorld
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     [self runTimer];
@@ -385,14 +424,14 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:test
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     _shouldBracket = NO;
     [self runTimer];
     NSString *expected = [[@"\x1b[200~" stringByAppendingString:test] stringByAppendingString:@"\x1b[201~"];
     XCTAssert([_writeBuffer isEqualToString:expected]);
-    XCTAssert(fabs(_helper.duration - 0.01) < kFloatingPointTolerance);
+    XCTAssert(fabs(_helper.duration - 0.02) < kFloatingPointTolerance);
 }
 
 - (void)testDelegateChangesItsMindAboutBracketingWithQueue {
@@ -401,7 +440,7 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:test1
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     _shouldBracket = NO;
@@ -410,7 +449,7 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:test2
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
 
@@ -424,12 +463,12 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:test
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     [self runTimer];
     XCTAssert([_writeBuffer isEqualToString:test]);
-    XCTAssert(fabs(_helper.duration - 0.01) < kFloatingPointTolerance);
+    XCTAssert(fabs(_helper.duration - 0.02) < kFloatingPointTolerance);
 }
 
 - (void)testSlowTwoChunkPasteString {
@@ -437,7 +476,7 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:test
                   slowly:YES
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     [self runTimer];
@@ -450,19 +489,19 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:test1
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     NSString *test2 = [@"2" stringRepeatedTimes:2000];
     [_helper pasteString:test2
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     [self runTimer];
     XCTAssert([_writeBuffer isEqualToString:[test1 stringByAppendingString:test2]]);
-    NSTimeInterval expectedDuration = 2 * 0.01;
+    NSTimeInterval expectedDuration = 4 * 0.01;
     XCTAssert(fabs(_helper.duration - expectedDuration) < kFloatingPointTolerance);
     XCTAssert([[[[PasteboardHistory sharedInstance] entries][0] mainValue] isEqualToString:test1]);
     XCTAssert([[[[PasteboardHistory sharedInstance] entries][1] mainValue] isEqualToString:test2]);
@@ -473,10 +512,10 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:test1
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
-    [_helper enqueueEvent:[NSEvent keyEventWithType:NSKeyDown
+    [_helper enqueueEvent:[NSEvent keyEventWithType:NSEventTypeKeyDown
                                            location:NSZeroPoint
                                       modifierFlags:0
                                           timestamp:[NSDate timeIntervalSinceReferenceDate]
@@ -490,12 +529,12 @@ static const double kFloatingPointTolerance = 0.00001;
     [_helper pasteString:test2
                   slowly:NO
         escapeShellChars:NO
-                commands:NO
+                isUpload:NO
             tabTransform:kTabTransformNone
             spacesPerTab:0];
     [self runTimer];
     XCTAssert([_writeBuffer isEqualToString:[[test1 stringByAppendingString:@"x"] stringByAppendingString:test2]]);
-    NSTimeInterval expectedDuration = 2 * 0.01;
+    NSTimeInterval expectedDuration = 4 * 0.01;
     XCTAssert(fabs(_helper.duration - expectedDuration) < kFloatingPointTolerance);
 }
 
@@ -533,6 +572,19 @@ static const double kFloatingPointTolerance = 0.00001;
     return NO;
 }
 
+- (void)pasteHelperPasteViewVisibilityDidChange {
+}
+
+- (iTermVariableScope *)pasteHelperScope {
+    return nil;
+}
+
+
+- (iTermStatusBarViewController *)pasteHelperStatusBarViewController {
+    return nil;
+}
+
+
 #pragma mark - iTermWarningHandler
 
 - (NSModalResponse)warningWouldShowAlert:(NSAlert *)alert identifier:(NSString *)identifier {
