@@ -11,8 +11,10 @@
 #import "PreferencePanel.h"
 #import "NSPopUpButton+iTerm.h"
 #import "iTermApplicationDelegate.h"
+#import "iTermPasteSpecialViewController.h"
 #import "ITAddressBookMgr.h"
 #import "FutureMethods.h"
+#import "NSTextField+iTerm.h"
 
 static NSString *kPointerActionsKey = @"PointerActions";  // Used in NSUserDefaults
 static NSString *kActionKey = @"Action";  // Used within values
@@ -72,13 +74,15 @@ NSString *kSelectNextPanePointerAction = @"kSelectNextPanePointerAction";
 NSString *kSelectPreviousPanePointerAction = @"kSelectPreviousPanePointerAction";
 NSString *kExtendSelectionPointerAction = @"kExtendSelectionPointerAction";
 NSString *kQuickLookAction = @"kQuickLookAction";
+NSString *kIgnoreAction = @"kIgnoreAction";
 
 typedef enum {
     kNoArg,
     kEscPlusArg,
     kHexCodeArg,
     kTextArg,
-    kProfileArg
+    kProfileArg,
+    kAdvancedPasteArg
 } ArgumentType;
 
 @interface NSString (PointerPrefsController)
@@ -103,7 +107,7 @@ typedef enum {
 + (NSString *)localizedNumClicks:(int)n;
 + (NSString *)localizedButtonNameForButtonNumber:(int)n;
 + (NSString *)localizedGestureNameForGestureIdentifier:(NSString *)ident;
-+ (NSString *)localizedModifers:(int)keyMods;
++ (NSString *)localizedModifiers:(int)keyMods;
 + (NSDictionary *)localizedActionMap;
 + (NSString *)localizedActionForDict:(NSDictionary *)dict;
 + (NSString *)localizedActionForKey:(NSString *)key;
@@ -143,15 +147,17 @@ typedef enum {
 
     IBOutlet NSButton *ok_;
     IBOutlet NSButton *remove_;
+    iTermPasteSpecialViewController *_pasteSpecialViewController;
+    IBOutlet NSView *_pasteSpecialViewContainer;
+    NSRect _initialFrame;
+    NSRect _initialPasteContainerFrame;
 
     NSString *origKey_;
 }
 
 - (void)dealloc {
-    [origKey_ release];
     tableView_.delegate = nil;
     tableView_.dataSource = nil;
-    [super dealloc];
 }
 
 + (NSDictionary *)dictForAction:(NSString *)action
@@ -162,16 +168,16 @@ typedef enum {
 + (NSString *)modCharsForMask:(int)modifiers
 {
     NSMutableString *modStr = [NSMutableString string];
-    if (modifiers & NSCommandKeyMask) {
+    if (modifiers & NSEventModifierFlagCommand) {
         [modStr appendString:kCommandKeyChar];
     }
-    if (modifiers & NSAlternateKeyMask) {
+    if (modifiers & NSEventModifierFlagOption) {
         [modStr appendString:kOptionKeyChar];
     }
-    if (modifiers & NSShiftKeyMask) {
+    if (modifiers & NSEventModifierFlagShift) {
         [modStr appendString:kShiftKeyChar];
     }
-    if (modifiers & NSControlKeyMask) {
+    if (modifiers & NSEventModifierFlagControl) {
         [modStr appendString:kControlKeyChar];
     }
     return modStr;
@@ -181,16 +187,16 @@ typedef enum {
 {
     int mask = 0;
     if ([modChars rangeOfString:kCommandKeyChar].location != NSNotFound) {
-        mask |= NSCommandKeyMask;
+        mask |= NSEventModifierFlagCommand;
     }
     if ([modChars rangeOfString:kOptionKeyChar].location != NSNotFound) {
-        mask |= NSAlternateKeyMask;
+        mask |= NSEventModifierFlagOption;
     }
     if ([modChars rangeOfString:kShiftKeyChar].location != NSNotFound) {
-        mask |= NSShiftKeyMask;
+        mask |= NSEventModifierFlagShift;
     }
     if ([modChars rangeOfString:kControlKeyChar].location != NSNotFound) {
-        mask |= NSControlKeyMask;
+        mask |= NSEventModifierFlagControl;
     }
     return mask;
 }
@@ -378,15 +384,16 @@ typedef enum {
     }
 }
 
-+ (NSString *)localizedModifers:(int)keyMods {
++ (NSString *)localizedModifiers:(int)keyMods {
     return [NSString stringForModifiersWithMask:keyMods];
 }
 
 + (NSDictionary *)localizedActionMap
 {
     NSDictionary *names = [NSDictionary dictionaryWithObjectsAndKeys:
-                           @"Paste from Clipboard", kPasteFromClipboardPointerAction,
-                           @"Paste from Selection", kPasteFromSelectionPointerAction,
+                           @"Ignore", kIgnoreAction,
+                           @"Paste from Clipboard…", kPasteFromClipboardPointerAction,
+                           @"Paste from Selection…", kPasteFromSelectionPointerAction,
                            @"Extend Selection", kExtendSelectionPointerAction,
                            @"Open URL/Semantic History", kOpenTargetPointerAction,
                            @"Open URL in background", kOpenTargetInBackgroundPointerAction,
@@ -420,13 +427,15 @@ typedef enum {
 + (ArgumentType)argumentTypeForAction:(NSString *)action
 {
     NSDictionary *args = [NSDictionary dictionaryWithObjectsAndKeys:
-                          [NSNumber numberWithInt:kEscPlusArg], kSendEscapeSequencePointerAction,
-                          [NSNumber numberWithInt:kHexCodeArg], kSendHexCodePointerAction,
-                          [NSNumber numberWithInt:kTextArg], kSendTextPointerAction,
-                          [NSNumber numberWithInt:kProfileArg], kNewWindowWithProfilePointerAction,
-                          [NSNumber numberWithInt:kProfileArg], kNewTabWithProfilePointerAction,
-                          [NSNumber numberWithInt:kProfileArg], kNewVerticalSplitWithProfilePointerAction,
-                          [NSNumber numberWithInt:kProfileArg], kNewHorizontalSplitWithProfilePointerAction,
+                          @(kEscPlusArg), kSendEscapeSequencePointerAction,
+                          @(kHexCodeArg), kSendHexCodePointerAction,
+                          @(kTextArg), kSendTextPointerAction,
+                          @(kProfileArg), kNewWindowWithProfilePointerAction,
+                          @(kProfileArg), kNewTabWithProfilePointerAction,
+                          @(kProfileArg), kNewVerticalSplitWithProfilePointerAction,
+                          @(kProfileArg), kNewHorizontalSplitWithProfilePointerAction,
+                          @(kAdvancedPasteArg), kPasteFromClipboardPointerAction,
+                          @(kAdvancedPasteArg), kPasteFromSelectionPointerAction,
                           nil];
     NSNumber *n = [args objectForKey:action];
     if (n) {
@@ -481,6 +490,13 @@ typedef enum {
                 return [name stringByReplacingOccurrencesOfString:@"…"
                                                        withString:[NSString stringWithFormat:@" \"%@\"", bookmarkName]];
             }
+            case kAdvancedPasteArg: {
+                if (argument.length) {
+                    name = [NSString stringWithFormat:@"%@: %@",
+                            [name stringByReplacingOccurrencesOfString:@"…" withString:@""],
+                            [iTermPasteSpecialViewController descriptionForCodedSettings:argument]];
+                }
+            }
         }
     }
 
@@ -533,18 +549,9 @@ typedef enum {
     static NSDictionary *defaultDict;
     if (!defaultDict) {
         NSMutableDictionary *temp = [NSMutableDictionary dictionaryWithDictionary:[PointerPrefsController defaultSettings]];
-        // Migrate old global prefs into the dict.
-        if (![[NSUserDefaults standardUserDefaults] boolForKey:@"PasteFromClipboard"]) {
-            NSDictionary *middleButtonPastesFromSelection =
-                [PointerPrefsController dictForAction:kPasteFromSelectionPointerAction];
-            [temp setObject:middleButtonPastesFromSelection
-                     forKey:[PointerPrefsController keyForButton:kMiddleButton
-                                                          clicks:1
-                                                       modifiers:0]];
-        }
         if ([iTermPreferences boolForKey:kPreferenceKeyThreeFingerEmulatesMiddle]) {
             // Find all actions that use middle button and add corresponding three-finger gesture.
-            NSMutableDictionary *tempCopy = [[temp mutableCopy] autorelease];
+            NSMutableDictionary *tempCopy = [temp mutableCopy];
             for (NSString *key in temp) {
                 if ([PointerPrefsController keyIsButton:key] &&
                     [PointerPrefsController buttonForKey:key] == kMiddleButton) {
@@ -555,7 +562,7 @@ typedef enum {
                 }
             }
             temp = tempCopy;
-            int modMasks[] = { NSCommandKeyMask, NSAlternateKeyMask, NSControlKeyMask, NSShiftKeyMask };
+            int modMasks[] = { NSEventModifierFlagCommand, NSEventModifierFlagOption, NSEventModifierFlagControl, NSEventModifierFlagShift };
             int numModCombos = 1 << (sizeof(modMasks) / sizeof(int));
             for (int numClicks = 0; numClicks <= kMaxClicks; numClicks++) {
                 // i is a bitmask over the modMasks array indices.
@@ -578,7 +585,7 @@ typedef enum {
                 }
             }
         }
-        defaultDict = [temp retain];
+        defaultDict = temp;
     }
     return defaultDict;
 }
@@ -629,7 +636,9 @@ typedef enum {
     NSString *key = [PointerPrefsController keyForButton:buttonNumber
                                                   clicks:numClicks
                                                modifiers:modMask];
-    return [[[PointerPrefsController settings] objectForKey:key] objectForKey:kArgumentKey];
+    NSDictionary *settings = [PointerPrefsController settings];
+    NSDictionary *setting = [settings objectForKey:key];
+    return [setting objectForKey:kArgumentKey];
 }
 
 + (NSString *)actionWithButton:(int)buttonNumber
@@ -640,7 +649,9 @@ typedef enum {
                                                   clicks:numClicks
                                                modifiers:modMask];
     DLog(@"Look up key %@", key);
-    NSString *action = [[[PointerPrefsController settings] objectForKey:key] objectForKey:kActionKey];
+    NSDictionary *settings = [PointerPrefsController settings];
+    NSDictionary *setting = [settings objectForKey:key];
+    NSString *action = [setting objectForKey:kActionKey];
     return action;
 }
 
@@ -675,7 +686,9 @@ typedef enum {
     key = [PointerPrefsController keyForGesture:gesture
                                       modifiers:modMask];
     DLog(@"Look up action for gesture %@", key);
-    return [[[PointerPrefsController settings] objectForKey:key] objectForKey:kActionKey];
+    NSDictionary *settings = [PointerPrefsController settings];
+    NSDictionary *setting = [settings objectForKey:key];
+    return [setting objectForKey:kActionKey];
 }
 
 + (NSString *)argumentForGesture:(NSString *)gesture
@@ -684,7 +697,9 @@ typedef enum {
     NSString *key;
     key = [PointerPrefsController keyForGesture:gesture
                                       modifiers:modMask];
-    return [[[PointerPrefsController settings] objectForKey:key] objectForKey:kArgumentKey];
+    NSDictionary *settings = [PointerPrefsController settings];
+    NSDictionary *setting = [settings objectForKey:key];
+    return [setting objectForKey:kArgumentKey];
 }
 
 + (BOOL)haveThreeFingerTapEvents
@@ -708,7 +723,7 @@ typedef enum {
 {
     NSString *button = [PointerPrefsController localizedButtonNameForButtonNumber:buttonNumber];
     NSString *numClicks = [PointerPrefsController localizedNumClicks:clicks];
-    NSString *modifiers = [PointerPrefsController localizedModifers:modFlags];
+    NSString *modifiers = [PointerPrefsController localizedModifiers:modFlags];
     if ([modifiers length]) {
         modifiers = [modifiers stringByAppendingString:@" + "];
     }
@@ -724,10 +739,10 @@ typedef enum {
 
 - (void)setModifierButtons:(int)modMask
 {
-    [editModifiersCommand_ setState:(modMask & NSCommandKeyMask) ? NSOnState : NSOffState];
-    [editModifiersOption_ setState:(modMask & NSAlternateKeyMask) ? NSOnState : NSOffState];
-    [editModifiersShift_ setState:(modMask & NSShiftKeyMask) ? NSOnState : NSOffState];
-    [editModifiersControl_ setState:(modMask & NSControlKeyMask) ? NSOnState : NSOffState];
+    [editModifiersCommand_ setState:(modMask & NSEventModifierFlagCommand) ? NSControlStateValueOn : NSControlStateValueOff];
+    [editModifiersOption_ setState:(modMask & NSEventModifierFlagOption) ? NSControlStateValueOn : NSControlStateValueOff];
+    [editModifiersShift_ setState:(modMask & NSEventModifierFlagShift) ? NSControlStateValueOn : NSControlStateValueOff];
+    [editModifiersControl_ setState:(modMask & NSEventModifierFlagControl) ? NSControlStateValueOn : NSControlStateValueOff];
 }
 
 - (void)setButtonNumber:(int)buttonNumber clickCount:(int)clickCount modifiers:(int)modMask
@@ -760,7 +775,7 @@ typedef enum {
         if (isButton) {
             return [PointerPrefsController localizedButtonKey:key];
         } else {
-            NSString *modifiers = [PointerPrefsController localizedModifers:[PointerPrefsController modifiersForKey:key]];
+            NSString *modifiers = [PointerPrefsController localizedModifiers:[PointerPrefsController modifiersForKey:key]];
             if ([modifiers length]) {
                 modifiers = [modifiers stringByAppendingString:@" + "];
             }
@@ -790,18 +805,22 @@ typedef enum {
     }
 }
 
-+ (NSString *)actionForKey:(NSString *)key
-{
-    return [[[PointerPrefsController settings] objectForKey:key] objectForKey:kActionKey];
++ (NSString *)actionForKey:(NSString *)key {
+    NSDictionary *setting = [[PointerPrefsController settings] objectForKey:key];
+    return [setting objectForKey:kActionKey];
 }
 
-+ (NSString *)argumentForKey:(NSString *)key
-{
-    return [[[PointerPrefsController settings] objectForKey:key] objectForKey:kArgumentKey];
++ (NSString *)argumentForKey:(NSString *)key {
+    NSDictionary *setting = [[PointerPrefsController settings] objectForKey:key];
+    return [setting objectForKey:kArgumentKey];
 }
 
 - (void)updateArgumentFieldsForAction:(NSString *)actionIdent argument:(NSString *)currentArg
 {
+    if (NSEqualRects(NSZeroRect, _initialFrame)) {
+        _initialFrame = _pasteSpecialViewContainer.window.frame;
+        _initialPasteContainerFrame = _pasteSpecialViewContainer.frame;
+    }
     ArgumentType argType = kNoArg;
     if (actionIdent) {
         argType = [PointerPrefsController argumentTypeForAction:actionIdent];
@@ -811,6 +830,7 @@ typedef enum {
             [editArgumentLabel_ setHidden:YES];
             [editArgumentField_ setHidden:YES];
             [editArgumentButton_ setHidden:YES];
+            _pasteSpecialViewContainer.hidden = YES;
             break;
 
         case kEscPlusArg:
@@ -823,6 +843,7 @@ typedef enum {
             [editArgumentField_ setStringValue:currentArg];
             [editArgumentField_ setRefusesFirstResponder:NO];
             [editArgumentField_ setSelectable:YES];
+            _pasteSpecialViewContainer.hidden = YES;
             break;
 
         case kHexCodeArg:
@@ -833,6 +854,7 @@ typedef enum {
             [editArgumentLabel_ setStringValue:@"Hex codes:"];
             [[editArgumentField_ cell] setPlaceholderString:@"ex: 0x7f 0x20"];
             [editArgumentField_ setStringValue:currentArg];
+            _pasteSpecialViewContainer.hidden = YES;
             break;
 
         case kTextArg:
@@ -843,6 +865,7 @@ typedef enum {
             [editArgumentLabel_ setStringValue:@"Text:"];
             [[editArgumentField_ cell] setPlaceholderString:@"Enter value to send"];
             [editArgumentField_ setStringValue:currentArg];
+            _pasteSpecialViewContainer.hidden = YES;
             break;
 
         case kProfileArg:
@@ -851,10 +874,18 @@ typedef enum {
             [editArgumentButton_ setHidden:NO];
             [editArgumentLabel_ setStringValue:@"Profile:"];
             [editArgumentButton_ populateWithProfilesSelectingGuid:currentArg];
-
+            _pasteSpecialViewContainer.hidden = YES;
             break;
 
+        case kAdvancedPasteArg:
+            editArgumentLabel_.hidden = YES;
+            editArgumentField_.hidden = YES;
+            editArgumentButton_.hidden = YES;
+            _pasteSpecialViewContainer.hidden = NO;
+            [self configurePasteSpecialWithArgument:currentArg];
+            break;
     }
+    [self updateWindowFrame];
 }
 
 - (void)loadKeyIntoEditPane:(NSString *)key
@@ -873,10 +904,10 @@ typedef enum {
     NSString *currentArg = [PointerPrefsController argumentForKey:key];
     [self updateArgumentFieldsForAction:actionIdent argument:currentArg];
 
-    [editModifiersCommand_ setState:(modMask & NSCommandKeyMask) ? NSOnState : NSOffState];
-    [editModifiersOption_ setState:(modMask & NSAlternateKeyMask) ? NSOnState : NSOffState];
-    [editModifiersShift_ setState:(modMask & NSShiftKeyMask) ? NSOnState : NSOffState];
-    [editModifiersControl_ setState:(modMask & NSControlKeyMask) ? NSOnState : NSOffState];
+    [editModifiersCommand_ setState:(modMask & NSEventModifierFlagCommand) ? NSControlStateValueOn : NSControlStateValueOff];
+    [editModifiersOption_ setState:(modMask & NSEventModifierFlagOption) ? NSControlStateValueOn : NSControlStateValueOff];
+    [editModifiersShift_ setState:(modMask & NSEventModifierFlagShift) ? NSControlStateValueOn : NSControlStateValueOff];
+    [editModifiersControl_ setState:(modMask & NSEventModifierFlagControl) ? NSControlStateValueOn : NSControlStateValueOff];
     [editAction_ selectItemWithTitle:localizedAction];
     BOOL isButton = !key || [PointerPrefsController keyIsButton:key];
     if (isButton) {
@@ -890,8 +921,7 @@ typedef enum {
         [editButton_ selectItemWithTag:[PointerPrefsController tagForGestureIdentifier:gestureIdent]];
         [editClickType_ selectItem:nil];
     }
-    [origKey_ autorelease];
-    origKey_ = [key retain];
+    origKey_ = key;
     [self buttonOrGestureChanged:nil];
     [ok_ setEnabled:[self okShouldBeEnabled]];
 }
@@ -899,10 +929,10 @@ typedef enum {
 - (IBAction)buttonOrGestureChanged:(id)sender
 {
     if ([editButton_ selectedTag] >= kMinGestureTag) {
-        [editClickTypeLabel_ setTextColor:[NSColor disabledControlTextColor]];
+        editClickTypeLabel_.labelEnabled = NO;
         [editClickType_ setEnabled:NO];
     } else {
-        [editClickTypeLabel_ setTextColor:[NSColor blackColor]];
+        editClickTypeLabel_.labelEnabled = YES;
         [editClickType_ setEnabled:YES];
     }
 }
@@ -921,10 +951,7 @@ typedef enum {
     self.hasSelection = [tableView_ numberOfSelectedRows] > 0;
     int rowIndex = [tableView_ selectedRow];
 
-    NSColor *textColor = [NSColor disabledControlTextColor];
     if (self.hasSelection) {
-        textColor = [NSColor blackColor];
-
         NSString *key = [PointerPrefsController keyForRowIndex:rowIndex];
         NSDictionary *action = [[PointerPrefsController settings] objectForKey:key];
 
@@ -932,14 +959,14 @@ typedef enum {
         [editAction_ selectItemWithTitle:[PointerPrefsController localizedActionForDict:action]];
 
         int modflags = [PointerPrefsController modifiersForKey:key];
-        [editModifiersCommand_ setState:(modflags & NSCommandKeyMask) ? NSOnState : NSOffState];
-        [editModifiersOption_ setState:(modflags & NSAlternateKeyMask) ? NSOnState : NSOffState];
-        [editModifiersShift_ setState:(modflags & NSShiftKeyMask) ? NSOnState : NSOffState];
-        [editModifiersControl_ setState:(modflags & NSControlKeyMask) ? NSOnState : NSOffState];
+        [editModifiersCommand_ setState:(modflags & NSEventModifierFlagCommand) ? NSControlStateValueOn : NSControlStateValueOff];
+        [editModifiersOption_ setState:(modflags & NSEventModifierFlagOption) ? NSControlStateValueOn : NSControlStateValueOff];
+        [editModifiersShift_ setState:(modflags & NSEventModifierFlagShift) ? NSControlStateValueOn : NSControlStateValueOff];
+        [editModifiersControl_ setState:(modflags & NSEventModifierFlagControl) ? NSControlStateValueOn : NSControlStateValueOff];
     }
-    [editButtonLabel_ setTextColor:textColor];
-    [editModifiersLabel_ setTextColor:textColor];
-    [editActionLabel_ setTextColor:textColor];
+    editButtonLabel_.labelEnabled = self.hasSelection;
+    editModifiersLabel_.labelEnabled = self.hasSelection;
+    editActionLabel_.labelEnabled = self.hasSelection;
 }
 
 - (void)tableViewRowDoubleClicked:(id)sender
@@ -950,23 +977,19 @@ typedef enum {
     }
 }
 
-- (void)editKey:(NSString *)key
-{
+- (void)editKey:(NSString *)key {
     [self loadKeyIntoEditPane:key];
-    [NSApp beginSheet:panel_
-       modalForWindow:[[PreferencePanel sharedInstance] window]
-        modalDelegate:self
-       didEndSelector:@selector(genericCloseSheet:returnCode:contextInfo:)
-          contextInfo:key];
+    __weak __typeof(self) weakSelf = self;
+    [[[PreferencePanel sharedInstance] window] beginSheet:panel_
+                                        completionHandler:^(NSModalResponse returnCode) {
+                                            __strong __typeof(weakSelf) strongSelf = self;
+                                            if (strongSelf) {
+                                                [strongSelf->panel_ close];
+                                            }
+                                        }];
 }
 
-- (void)genericCloseSheet:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo
-{
-    [sheet close];
-}
-
-- (IBAction)ok:(id)sender
-{
+- (IBAction)ok:(id)sender {
     NSMutableDictionary *temp = [NSMutableDictionary dictionaryWithDictionary:[PointerPrefsController settings]];
     if (origKey_) {
         [temp removeObjectForKey:origKey_];
@@ -990,20 +1013,25 @@ typedef enum {
             [newValue setObject:[[editArgumentButton_ selectedItem] title]
                          forKey:kArgumentKey];
         }
+    } else if (!_pasteSpecialViewContainer.isHidden) {
+        if ([PointerPrefsController argumentTypeForAction:theAction] == kAdvancedPasteArg) {
+            [newValue setObject:[_pasteSpecialViewController stringEncodedSettings] ?: @""
+                         forKey:kArgumentKey];
+        }
     }
     NSString *newKey;
     int modMask = 0;
-    if ([editModifiersCommand_ state] == NSOnState) {
-        modMask |= NSCommandKeyMask;
+    if ([editModifiersCommand_ state] == NSControlStateValueOn) {
+        modMask |= NSEventModifierFlagCommand;
     }
-    if ([editModifiersOption_ state] == NSOnState) {
-        modMask |= NSAlternateKeyMask;
+    if ([editModifiersOption_ state] == NSControlStateValueOn) {
+        modMask |= NSEventModifierFlagOption;
     }
-    if ([editModifiersShift_ state] == NSOnState) {
-        modMask |= NSShiftKeyMask;
+    if ([editModifiersShift_ state] == NSControlStateValueOn) {
+        modMask |= NSEventModifierFlagShift;
     }
-    if ([editModifiersControl_ state] == NSOnState) {
-        modMask |= NSControlKeyMask;
+    if ([editModifiersControl_ state] == NSControlStateValueOn) {
+        modMask |= NSEventModifierFlagControl;
     }
     if ([editButton_ selectedTag] >= kMinGestureTag) {
         // Gesture
@@ -1018,12 +1046,11 @@ typedef enum {
     [temp setObject:newValue forKey:newKey];
     [PointerPrefsController setSettings:temp];
     [tableView_ reloadData];
-    [NSApp endSheet:panel_];
+    [[[PreferencePanel sharedInstance] window] endSheet:panel_];
 }
 
-- (IBAction)cancel:(id)sender
-{
-    [NSApp endSheet:panel_];
+- (IBAction)cancel:(id)sender {
+    [[[PreferencePanel sharedInstance] window] endSheet:panel_];
 }
 
 - (IBAction)add:(id)sender
@@ -1063,6 +1090,41 @@ typedef enum {
     [tableView_ reloadData];
 }
 
+- (void)configurePasteSpecialWithArgument:(NSString *)parameterValue {
+    _pasteSpecialViewController = [[iTermPasteSpecialViewController alloc] init];
+    [_pasteSpecialViewController view];
+
+    // Set a few defaults; otherwise everything is reasonable.
+    _pasteSpecialViewController.numberOfSpacesPerTab = [iTermPreferences intForKey:kPreferenceKeyPasteSpecialSpacesPerTab];
+    _pasteSpecialViewController.shouldRemoveNewlines = NO;
+    _pasteSpecialViewController.shouldBase64Encode = NO;
+    _pasteSpecialViewController.shouldWaitForPrompt = NO;
+    _pasteSpecialViewController.shouldEscapeShellCharsWithBackslash = NO;
+    if (parameterValue.length > 0) {
+        [_pasteSpecialViewController loadSettingsFromString:parameterValue];
+    }
+    _pasteSpecialViewController.view.frame = _pasteSpecialViewController.view.bounds;
+    NSRect theFrame = _pasteSpecialViewContainer.frame;
+    CGFloat originalHeight = theFrame.size.height;
+    theFrame.size = _pasteSpecialViewController.view.bounds.size;
+    theFrame.origin.y -= (theFrame.size.height - originalHeight);
+    _pasteSpecialViewContainer.frame = theFrame;
+    [_pasteSpecialViewContainer addSubview:_pasteSpecialViewController.view];
+}
+
+- (void)updateWindowFrame {
+    NSRect frame;
+    if (_pasteSpecialViewContainer.isHidden) {
+        frame = _initialFrame;
+    } else {
+        frame = _initialFrame;
+        NSSize desiredSize = _pasteSpecialViewController.view.frame.size;
+        frame.size.width += desiredSize.width - _initialPasteContainerFrame.size.width;
+        frame.size.height += desiredSize.height - _initialPasteContainerFrame.size.height;
+    }
+    [_pasteSpecialViewContainer.window setFrame:frame display:YES animate:YES];
+}
+
 @end
 
 @implementation NSString (PointerPrefsController)
@@ -1137,5 +1199,3 @@ typedef enum {
 }
 
 @end
-
-
