@@ -13,13 +13,17 @@
 #import <NMSSH/libssh2.h>
 
 #import "DebugLogging.h"
+#import "ITAddressBookMgr.h"
+#import "iTermSlowOperationGateway.h"
+#import "iTermOpenDirectory.h"
 #import "iTermWarning.h"
 #import "NSFileManager+iTerm.h"
 #import "NSObject+iTerm.h"
-#import "NSStringiTerm.h"
+#import "NSStringITerm.h"
+#import "ProfileModel.h"
 
 @interface NMSSHSession(iTerm)
-- (id)agent;
+@property (atomic, readonly) void *agent;
 @end
 
 static NSString *const kSCPFileErrorDomain = @"com.googlecode.iterm2.SCPFile";
@@ -31,8 +35,65 @@ static NSError *SCPFileError(NSString *description) {
                            userInfo:@{ NSLocalizedDescriptionKey: description }];
 }
 
+@interface iTermAuthSock: NSObject
+@property (nonatomic, readonly) NSString *authSock;
++ (instancetype)sharedInstance;
+@end
+
+@implementation iTermAuthSock {
+    NSString *_authSock;
+    dispatch_group_t _group;
+}
+
++ (instancetype)sharedInstance {
+    static iTermAuthSock *instance;
+    static dispatch_once_t onceToken;
+    dispatch_once(&onceToken, ^{
+        instance = [[self alloc] init];
+    });
+    return instance;
+}
+
+- (NSString *)authSock {
+    Profile *profile = [[ProfileModel sharedInstance] defaultBookmark];
+    NSString *shell = [ITAddressBookMgr customShellForProfile:profile] ?: [iTermOpenDirectory userShell] ?:  @"/bin/bash";
+    dispatch_group_t group;
+    BOOL request = NO;
+    @synchronized(self) {
+        if (_authSock) {
+            return _authSock;
+        }
+        if (!_group) {
+            _group = dispatch_group_create();
+            request = YES;
+        }
+        group = _group;
+    }
+    if (request) {
+        dispatch_group_enter(group);
+        DLog(@"Try to get the value of $SSH_AUTH_SOCK");
+        [[iTermSlowOperationGateway sharedInstance] exfiltrateEnvironmentVariableNamed:@"SSH_AUTH_SOCK"
+                                                                                 shell:shell
+                                                                            completion:^(NSString * _Nonnull value) {
+            @synchronized(self) {
+                self->_authSock = value;
+            }
+            dispatch_group_leave(group);
+            DLog(@"Value is %@", value);
+        }];
+    }
+    dispatch_group_wait(group, dispatch_time(DISPATCH_TIME_NOW,
+                                             0.5 * NSEC_PER_SEC));
+    @synchronized(self) {
+        DLog(@"Return %@", _authSock);
+        return _authSock;
+    }
+}
+
+@end
+
 @interface SCPFile () <NMSSHSessionDelegate>
-@property(atomic, assign) NMSSHSession *session;
+@property(atomic, strong) NMSSHSession *session;
 @property(atomic, assign) BOOL stopped;
 @property(atomic, copy) NSString *error;
 @property(atomic, copy) NSString *destination;
@@ -60,16 +121,6 @@ static NSError *SCPFileError(NSString *description) {
     return self;
 }
 
-- (void)dealloc {
-    [_error release];
-    [_destination release];
-    dispatch_release(_queue);
-    [_homeDirectory release];
-    [_userName release];
-    [_hostName release];
-    [super dealloc];
-}
-
 - (NSError *)lastError {
   if (self.session.rawSession) {
     return self.session.lastError;
@@ -82,11 +133,7 @@ static NSError *SCPFileError(NSString *description) {
 - (void)setQueue:(dispatch_queue_t)queue {
     @synchronized(self) {
         if (queue != _queue) {
-            dispatch_release(_queue);
             _queue = queue;
-            if (queue) {
-                dispatch_retain(queue);
-            }
         }
     }
 }
@@ -137,7 +184,7 @@ static NSError *SCPFileError(NSString *description) {
 - (NSString *)hostname {
     NSArray *hostComponents = [self.path.hostname componentsSeparatedByString:@":"];
     NSInteger components = [hostComponents count];
-    
+
     // Check if the host is {hostname}:{port} or {IPv4}:{port}
     if (components == 2) {
         return hostComponents[0];
@@ -149,30 +196,30 @@ static NSError *SCPFileError(NSString *description) {
         NSString *bracketedHostname = [hostComponents componentsJoinedByString:@":"];
         return [bracketedHostname substringWithRange:NSMakeRange(1, bracketedHostname.length - 2)];
     }
-    
+
     return self.path.hostname;
 }
 
 - (int)port {
     NSArray *hostComponents = [self.path.hostname componentsSeparatedByString:@":"];
     NSInteger components = [hostComponents count];
-    
+
     // Check if the host is {hostname}:{port} or {IPv4}:{port}
     if (components == 2) {
-        NSNumberFormatter *formatter = [[[NSNumberFormatter alloc] init] autorelease];
-        [formatter setLocale:[[[NSLocale alloc] initWithLocaleIdentifier:@"en_US"] autorelease]];
-        
+        NSNumberFormatter *formatter = [[NSNumberFormatter alloc] init];
+        [formatter setLocale:[[NSLocale alloc] initWithLocaleIdentifier:@"en_US"]];
+
         return [[formatter numberFromString:[hostComponents lastObject]] intValue];
     } else if (components >= 4 &&
                [hostComponents[0] hasPrefix:@"["] &&
                [hostComponents[components-2] hasSuffix:@"]"]) {
         // Check if the host is [{IPv6}]:{port}
-        NSNumberFormatter *formatter = [[[NSNumberFormatter alloc] init] autorelease];
-        [formatter setLocale:[[[NSLocale alloc] initWithLocaleIdentifier:@"en_US"] autorelease]];
-        
+        NSNumberFormatter *formatter = [[NSNumberFormatter alloc] init];
+        [formatter setLocale:[[NSLocale alloc] initWithLocaleIdentifier:@"en_US"]];
+
         return [[formatter numberFromString:[hostComponents lastObject]] intValue];
     }
-    
+
     // If no port was defined, use 22 by default
     return 22;
 }
@@ -205,7 +252,7 @@ static NSError *SCPFileError(NSString *description) {
             if (config) {
                 [configs addObject:config];
             } else {
-                ELog(@"Could not parse config file at %@", path);
+                XLog(@"Could not parse config file at %@", path);
             }
         }
     }
@@ -213,7 +260,7 @@ static NSError *SCPFileError(NSString *description) {
 }
 
 // This runs in a thread
-- (NSString *)filenameByExpandingMetasyntaticVariables:(NSString *)filename {
+- (NSString *)filenameByExpandingMetasyntacticVariables:(NSString *)filename {
     filename = [filename stringByExpandingTildeInPath];
     NSDictionary *substitutions =
         @{ @"%d": _homeDirectory,
@@ -232,6 +279,35 @@ static NSError *SCPFileError(NSString *description) {
     [self performTransfer:isDownload agentAllowed:YES];
 }
 
+// Don't call this on the main thread!
+- (NSString *)keyboardInteractiveRequest:(NSString *)prompt {
+    __block NSString *value = nil;
+    dispatch_group_t group = dispatch_group_create();
+    dispatch_group_enter(group);
+    dispatch_async(dispatch_get_main_queue(), ^() {
+        [[FileTransferManager sharedInstance] transferrableFile:self
+                                              interactivePrompt:prompt
+                                                     completion:^(NSString *result) {
+                                                         value = [result copy];
+                                                         dispatch_group_leave(group);
+                                                     }];
+    });
+    dispatch_group_wait(group, DISPATCH_TIME_FOREVER);
+
+    return value;
+}
+
+- (NSURL *)sessionURL {
+    assert(self.session);
+    NSURLComponents *components = [[NSURLComponents alloc] init];
+    components.host = self.session.host;
+    components.user = self.session.username;
+    components.port = self.session.port;
+    components.path = self.path.path;
+    components.scheme = @"ssh";
+    return components.URL;
+}
+
 // This runs in a thread.
 - (void)performTransfer:(BOOL)isDownload agentAllowed:(BOOL)agentAllowed {
     NSString *baseName = [[self class] fileNameForPath:self.path.path];
@@ -244,23 +320,27 @@ static NSError *SCPFileError(NSString *description) {
         return;
     }
     _okToAdd = NO;
+    int effectivePort;
     if (self.session) {
         self.session.delegate = self;
+        effectivePort = self.session.port.intValue;
     } else {
-        self.session = [[[NMSSHSession alloc] initWithHost:[self hostname]
-                                                   configs:[self configs]
-                                           withDefaultPort:[self port]
-                                           defaultUsername:self.path.username] autorelease];
+        self.session = [[NMSSHSession alloc] initWithHost:[self hostname]
+                                                  configs:[self configs]
+                                          withDefaultPort:[self port]
+                                          defaultUsername:self.path.username];
+        effectivePort = self.session.port.intValue;
         self.session.delegate = self;
         [self.session connect];
         if (self.stopped) {
-            ELog(@"Stop after connect");
+            XLog(@"Stop after connect");
             dispatch_sync(dispatch_get_main_queue(), ^() {
                 [[FileTransferManager sharedInstance] transferrableFileDidStopTransfer:self];
             });
             return;
         }
     }
+    NSURL *url = [self sessionURL];
     if (!self.session.isConnected) {
         NSError *theError = [self lastError];
         if (!theError) {
@@ -276,14 +356,15 @@ static NSError *SCPFileError(NSString *description) {
             [[FileTransferManager sharedInstance] transferrableFile:self
                                      didFinishTransmissionWithError:theError];
             iTermWarningSelection selection =
-                [iTermWarning showWarningWithTitle:[NSString stringWithFormat:@"Failed to connect to %@:%d. Double-check that the host name is correct.", self.hostname, self.port]
+                [iTermWarning showWarningWithTitle:[NSString stringWithFormat:@"Failed to connect to %@:%d. Double-check that the host name is correct.", self.hostname, effectivePort]
                                            actions:@[ @"Ok", @"Help" ]
                                      actionMapping:nil
                                          accessory:nil
                                         identifier:kSecureCopyConnectionFailedWarning
                                        silenceable:kiTermWarningTypePermanentlySilenceable
                                            heading:@"Connection Failed"
-                                       cancelLabel:@"Help"];
+                                       cancelLabel:@"Help"
+                                            window:nil];
             if (selection == kiTermWarningSelection1) {
                 [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"https://iterm2.com/troubleshoot-hostname"]];
             }
@@ -292,7 +373,8 @@ static NSError *SCPFileError(NSString *description) {
     }
 
     BOOL didConnectToAgent = NO;
-    if (agentAllowed) {
+    if (agentAllowed && !self.hasPredecessor) {
+        self.session.authSock = [[iTermAuthSock sharedInstance] authSock];
         [self.session connectToAgent];
         // Check a private property to see if the connection to the agent was made.
         if ([self.session respondsToSelector:@selector(agent)]) {
@@ -307,19 +389,15 @@ static NSError *SCPFileError(NSString *description) {
         }
         for (NSString *authType in authTypes) {
             if (self.stopped) {
-                ELog(@"Break out of auth loop because stopped");
+                XLog(@"Break out of auth loop because stopped");
                 break;
             }
             if (!self.session.session) {
-                ELog(@"Break out of auth loop because disconnected");
+                XLog(@"Break out of auth loop because disconnected");
                 break;
             }
             if ([authType isEqualToString:@"password"]) {
-                __block NSString *password;
-                dispatch_sync(dispatch_get_main_queue(), ^() {
-                    password = [[FileTransferManager sharedInstance] transferrableFile:self
-                                                             keyboardInteractivePrompt:@"password"];
-                });
+                NSString *password = [self keyboardInteractiveRequest:@"password"];
                 if (self.stopped || !password) {
                     break;
                 }
@@ -329,12 +407,7 @@ static NSError *SCPFileError(NSString *description) {
                 }
             } else if ([authType isEqualToString:@"keyboard-interactive"]) {
                 [self.session authenticateByKeyboardInteractiveUsingBlock:^NSString *(NSString *request) {
-                    __block NSString *response;
-                    dispatch_sync(dispatch_get_main_queue(), ^() {
-                        response = [[FileTransferManager sharedInstance] transferrableFile:self
-                                                                 keyboardInteractivePrompt:request];
-                    });
-                    return response;
+                    return [self keyboardInteractiveRequest:request];
                 }];
                 if (self.stopped || self.session.isAuthorized) {
                     break;
@@ -343,7 +416,7 @@ static NSError *SCPFileError(NSString *description) {
                 if (self.stopped) {
                     break;
                 }
-                
+
                 NSMutableArray *keyPaths = [NSMutableArray array];
                 if (self.session.hostConfig.identityFiles.count) {
                     [keyPaths addObjectsFromArray:self.session.hostConfig.identityFiles];
@@ -353,39 +426,35 @@ static NSError *SCPFileError(NSString *description) {
                                                      @"~/.ssh/id_ecdsa" ]];
                 }
                 NSFileManager *fileManager = [NSFileManager defaultManager];
-                for (NSString *keyPath in keyPaths) {
-                    keyPath = [self filenameByExpandingMetasyntaticVariables:keyPath];
+                for (NSString *iteratedKeyPath in keyPaths) {
+                    NSString *keyPath = [self filenameByExpandingMetasyntacticVariables:iteratedKeyPath];
                     if (![fileManager fileExistsAtPath:keyPath]) {
-                        ELog(@"No key file at %@", keyPath);
+                        XLog(@"No key file at %@", keyPath);
                         continue;
                     }
-                    __block NSString *password = nil;
+                    NSString *password = nil;
                     if ([self privateKeyIsEncrypted:keyPath]) {
-                        dispatch_sync(dispatch_get_main_queue(), ^() {
-                            NSString *prompt =
-                                [NSString stringWithFormat:@"passphrase for private key “%@”:",
-                                    keyPath];
-                            password = [[FileTransferManager sharedInstance] transferrableFile:self
-                                                                     keyboardInteractivePrompt:prompt];
-                        });
+                        NSString *prompt = [NSString stringWithFormat:@"passphrase for private key “%@”:",
+                                            keyPath];
+                        password = [self keyboardInteractiveRequest:prompt];
                     }
-                    ELog(@"Attempting to authenticate with key %@", keyPath);
+                    XLog(@"Attempting to authenticate with key %@", keyPath);
                     NSString *publicKeyPath = [keyPath stringByAppendingString:@".pub"];
                     if (![[NSFileManager defaultManager] fileExistsAtPath:publicKeyPath]) {
-                        ELog(@"Warning: no public key at %@. Trying to authenticate with only a private key.", publicKeyPath);
+                        XLog(@"Warning: no public key at %@. Trying to authenticate with only a private key.", publicKeyPath);
                         publicKeyPath = nil;
                     }
                     [self.session authenticateByPublicKey:publicKeyPath
                                                privateKey:keyPath
                                               andPassword:password];
-                
+
                     if (self.session.isAuthorized) {
-                        ELog(@"Authorized!");
+                        XLog(@"Authorized!");
                         break;
                     }
 
                     if (!self.session.session) {
-                        ELog(@"Disconnected!");
+                        XLog(@"Disconnected!");
                         break;
                     }
                 }
@@ -396,7 +465,7 @@ static NSError *SCPFileError(NSString *description) {
         }
     }
     if (self.stopped) {
-        ELog(@"Stop after auth");
+        XLog(@"Stop after auth");
         dispatch_sync(dispatch_get_main_queue(), ^() {
             [[FileTransferManager sharedInstance] transferrableFileDidStopTransfer:self];
         });
@@ -423,7 +492,7 @@ static NSError *SCPFileError(NSString *description) {
         });
         return;
     }
-    
+
     if (_okToAdd) {
         [self.session addKnownHostName:self.session.host
                                   port:[self.session.port intValue]
@@ -449,30 +518,48 @@ static NSError *SCPFileError(NSString *description) {
         }
         self.destination = tempfile;
         self.status = kTransferrableFileStatusTransferring;
+        __block BOOL quarantined = NO;
+        __block BOOL quarantineError = NO;
         BOOL ok = [self.session.channel downloadFile:self.path.path
                                                   to:tempfile
                                             progress:^BOOL (NSUInteger bytes, NSUInteger fileSize) {
-                                                self.bytesTransferred = bytes;
-                                                self.fileSize = fileSize;
-                                                dispatch_sync(dispatch_get_main_queue(), ^() {
-                                                    if (!self.stopped) {
-                                                        [[FileTransferManager sharedInstance] transferrableFileProgressDidChange:self];
-                                                    }
-                                                });
-                                                if (self.stopped) {
-                                                    ELog(@"Stopping mid-download");
-                                                }
-                                                return !self.stopped;
-                                            }];
-        __block NSError *error;
+            if (!quarantined) {
+                if (![self quarantine:tempfile sourceURL:url]) {
+                    quarantineError = YES;
+                    return NO;
+                }
+                quarantined = YES;
+            }
+            self.bytesTransferred = bytes;
+            self.fileSize = fileSize;
+            dispatch_sync(dispatch_get_main_queue(), ^() {
+                if (!self.stopped) {
+                    [[FileTransferManager sharedInstance] transferrableFileProgressDidChange:self];
+                }
+            });
+            if (self.stopped) {
+                XLog(@"Stopping mid-download");
+            }
+            return !self.stopped;
+        }];
+        if (!quarantined && [[NSFileManager defaultManager] fileExistsAtPath:tempfile]) {
+            // Zero-byte file, presumably.
+            if (![self quarantine:tempfile sourceURL:url]) {
+                quarantineError = YES;
+                ok = NO;
+            } else {
+                quarantined = YES;
+            }
+        }
+        __block NSError *error = nil;
         __block NSString *finalDestination = nil;
         if (ok) {
             error = nil;
             // We determine the filename and perform the move in the main thread to avoid two
             // threads trying to determine the final destination at the same time.
             dispatch_sync(dispatch_get_main_queue(), ^() {
-                finalDestination = [[self finalDestinationForPath:baseName
-                                             destinationDirectory:downloadDirectory] retain];
+                finalDestination = [self finalDestinationForPath:baseName
+                                            destinationDirectory:downloadDirectory];
                 [[NSFileManager defaultManager] moveItemAtPath:tempfile
                                                         toPath:finalDestination
                                                          error:&error];
@@ -482,20 +569,29 @@ static NSError *SCPFileError(NSString *description) {
                               tempfile, finalDestination];
             }
             [[NSFileManager defaultManager] removeItemAtPath:tempfile error:NULL];
-            self.destination = [finalDestination autorelease];
+            self.destination = finalDestination;
         } else {
-            [[NSFileManager defaultManager] removeItemAtPath:tempfile error:NULL];
+            const BOOL ok = [[NSFileManager defaultManager] removeItemAtPath:tempfile error:&error];
+            if (quarantineError && (!ok || error)) {
+                dispatch_sync(dispatch_get_main_queue(), ^{
+                    [self failedToRemoveUnquarantinedFileAt:tempfile];
+                });
+            }
             if (self.stopped) {
                 dispatch_sync(dispatch_get_main_queue(), ^() {
                     [[FileTransferManager sharedInstance] transferrableFileDidStopTransfer:self];
                 });
                 return;
             } else {
-                NSString *errorDescription = [[self lastError] localizedDescription];
-                if (errorDescription.length) {
-                    self.error = errorDescription;
+                if (quarantineError) {
+                    self.error = @"Quarantine Error";
                 } else {
-                    self.error = @"Download failed";
+                    NSString *errorDescription = [[self lastError] localizedDescription];
+                    if (errorDescription.length) {
+                        self.error = errorDescription;
+                    } else {
+                        self.error = @"Download failed";
+                    }
                 }
                 error = SCPFileError(@"Download failed");
             }
@@ -588,7 +684,7 @@ static NSError *SCPFileError(NSString *description) {
     self.fileSize = [[[NSFileManager defaultManager] attributesOfItemAtPath:self.localPath error:nil] fileSize];
     [[[FileTransferManager sharedInstance] files] addObject:self];
     [[FileTransferManager sharedInstance] transferrableFileDidStartTransfer:self];
-    
+
     if (!self.hasPredecessor) {
         dispatch_async(_queue, ^() {
             [self performTransferWrapper:NO];
@@ -615,11 +711,11 @@ static NSError *SCPFileError(NSString *description) {
             case NMSSHKnownHostStatusFailure:
                 title = [NSString stringWithFormat:@"Problem connecting to %@", session.host];
                 message = [NSString stringWithFormat:@"Could not read the known_hosts file.\n"
-                                                     @"As a result, the autenticity of host '%@' can't be established."
+                                                     @"As a result, the authenticity of host '%@' can't be established."
                                                      @"DSA key fingerprint is %@. Connect anyway?",
                            session.host, fingerprint];
                 break;
-                
+
             case NMSSHKnownHostStatusMatch:
                 result = YES;
                 message = nil;
@@ -634,12 +730,12 @@ static NSError *SCPFileError(NSString *description) {
                                                @"It is also possible that a host key has just been changed.\nConnect anyway?",
                      session.host, fingerprint];
                 break;
-                
+
             case NMSSHKnownHostStatusNotFound:
                 title = [NSString stringWithFormat:@"First time connecting to %@", session.host];
                 message =
                     [NSString stringWithFormat:@"The authenticity of host '%@' can't be established.\n\n"
-                                               @"DSA key fingerprint is %@.\n\nConnect anyay?",
+                                               @"DSA key fingerprint is %@.\n\nConnect anyway?",
                         session.host, fingerprint];
                 _okToAdd = YES;
                 break;
@@ -656,8 +752,7 @@ static NSError *SCPFileError(NSString *description) {
 - (NSString *)session:(NMSSHSession *)session keyboardInteractiveRequest:(NSString *)request {
     __block NSString *string;
     dispatch_sync(dispatch_get_main_queue(), ^() {
-        string = [[FileTransferManager sharedInstance] transferrableFile:self
-                                               keyboardInteractivePrompt:request];
+        string = [self keyboardInteractiveRequest:request];
     });
     return string;
 }
