@@ -20,12 +20,13 @@
 #import "VT100DCSParser.h"
 #import "DebugLogging.h"
 #import "NSStringITerm.h"
+#import "VT100SixelParser.h"
 #import "VT100StateMachine.h"
 #import "VT100TmuxParser.h"
 
 // Caps the amount of data to accumulate in _data before returning to the ground state. Prevents
 // a random ESC P from eating output forever by leaving us in the passthrough state until we get
-// an ST.
+// an ST. Note that there's an exception for file downloads wrapped in DCS tmux; … ST
 static const NSUInteger kMaxDataLength = 1024 * 1024;
 
 // Creates a unique-enough encoding of a DCS sequence at compile time so it can
@@ -54,8 +55,9 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     // |_hook|. This is here only for the tmux workaround (and the tmux parser is the only hook).
     BOOL _hookFinished;
 
-    // Holds the current state and the fine state machine.
-    VT100StateMachine *_stateMachine;
+    // Holds the current state and the fine state machine. Up to one of these will be nonnull.
+    VT100StateMachine *_stateMachineWithout8BitControlCharacters;
+    VT100StateMachine *_stateMachineWith8BitControlCharacters;
 
     // Concatenation of passthrough characters. Not added to when a hook is present.
     NSMutableString *_data;
@@ -71,6 +73,12 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     // is for the buggy tmux protocol, which is not terminated by ST alone (only %exit followed by
     // ST).
     id<VT100DCSParserHook> _hook;
+
+    BOOL _support8BitControlCharacters;
+
+    // We periodically check if _data looks like binary garbage. To avoid quadratic runtime we
+    // remember the number of UTF-16 codepoints that have been checked already.
+    NSInteger _checkedCount;
 }
 
 + (NSDictionary *)termcapTerminfoNameDictionary {
@@ -101,25 +109,25 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     return self;
 }
 
-- (void)dealloc {
-    [_stateMachine release];
-    [_data release];
-    [_parameterString release];
-    [_intermediateString release];
-    [_privateMarkers release];
-    [_hook release];
-    [super dealloc];
-}
-
 // Defines the state machine.
 - (VT100StateMachine *)stateMachine {
-    if (_stateMachine) {
-        return _stateMachine;
+    if (_support8BitControlCharacters) {
+        if (!_stateMachineWith8BitControlCharacters) {
+            _stateMachineWith8BitControlCharacters = [self newStateMachineWith8BitControlCharacters:YES];
+            _stateMachineWithout8BitControlCharacters = nil;
+        }
+        return _stateMachineWith8BitControlCharacters;
     }
 
-    VT100StateMachine *stateMachine;
+    if (!_stateMachineWithout8BitControlCharacters) {
+        _stateMachineWithout8BitControlCharacters = [self newStateMachineWith8BitControlCharacters:YES];
+        _stateMachineWith8BitControlCharacters = nil;
+    }
+    return _stateMachineWithout8BitControlCharacters;
+}
 
-    stateMachine = [[VT100StateMachine alloc] init];
+- (VT100StateMachine *)newStateMachineWith8BitControlCharacters:(BOOL)support8BitControlCharacters {
+    VT100StateMachine *stateMachine = [[VT100StateMachine alloc] init];
 
     // This is where we start and end.
     VT100State *groundState = [VT100State stateWithName:@"ground"
@@ -164,16 +172,22 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     [stateMachine addState:dcsEscapeState];
     [stateMachine addState:dcsPassthroughState];
 
+    __weak __typeof(self) weakSelf = self;
     for (VT100State *state in stateMachine.states) {
         [state addStateTransitionForCharacter:VT100CC_CAN
                                            to:groundState
-                                   withAction:^(unsigned char c) { [self execute]; } ];
+                                   withAction:^(unsigned char c) { [weakSelf execute]; } ];
         [state addStateTransitionForCharacter:VT100CC_SUB
                                            to:groundState
-                                   withAction:^(unsigned char c) { [self execute]; } ];
+                                   withAction:^(unsigned char c) { [weakSelf execute]; } ];
         [state addStateTransitionForCharacter:VT100CC_ESC
                                            to:dcsEscapeState
                                    withAction:nil ];
+        if (support8BitControlCharacters) {
+            [state addStateTransitionForCharacterRange:MakeCharacterRange(VT100CC_C1_IND, VT100CC_C1_APC)
+                                                    to:groundState
+                                            withAction:^(unsigned char c) { [weakSelf execute]; } ];
+        }
     }
 
     ////////////////////////////////////////////////////////////////////////////////////////////////
@@ -181,6 +195,11 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     [groundState addStateTransitionForCharacter:VT100CC_ESC
                                              to:escapeState
                                      withAction:nil];
+    if (support8BitControlCharacters) {
+        [groundState addStateTransitionForCharacter:VT100CC_C1_DCS
+                                                 to:dcsEntryState
+                                         withAction:nil];
+    }
 
     ////////////////////////////////////////////////////////////////////////////////////////////////
     // Add transitions from escape state.
@@ -197,12 +216,16 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     ////////////////////////////////////////////////////////////////////////////////////////////////
     // Add transitions from entry state.
     dcsEntryState.entryAction = ^(unsigned char c) {
-        _malformed = NO;
-        _executed = NO;
-        [_data setString:@""];
-        [_parameterString setString:@""];
-        [_intermediateString setString:@""];
-        [_privateMarkers setString:@""];
+        __strong __typeof(self) strongSelf = weakSelf;
+        if (strongSelf) {
+            strongSelf->_malformed = NO;
+            strongSelf->_executed = NO;
+            [strongSelf->_data setString:@""];
+            strongSelf->_checkedCount = 0;
+            [strongSelf->_parameterString setString:@""];
+            [strongSelf->_intermediateString setString:@""];
+            [strongSelf->_privateMarkers setString:@""];
+        }
     };
     // Got initial passthrough. Entry action will save it.
     [dcsEntryState addStateTransitionForCharacterRange:MakeCharacterRange('@', '~')
@@ -212,19 +235,28 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     [dcsEntryState addStateTransitionForCharacterRange:MakeCharacterRange(' ', '/')
                                                     to:dcsIntermediateState
                                             withAction:^(unsigned char c) {
-                                                [_intermediateString appendCharacter:c];
+                                                __strong __typeof(self) strongSelf = weakSelf;
+                                                if (strongSelf) {
+                                                    [strongSelf->_intermediateString appendCharacter:c];
+                                                }
                                             }];
     // Transition to parameter string.
     [dcsEntryState addStateTransitionForCharacterRange:MakeCharacterRange('0', '9')
                                                     to:dcsParamState
                                             withAction:^(unsigned char c) {
-                                                [_parameterString appendCharacter:c];
+                                                __strong __typeof(self) strongSelf = weakSelf;
+                                                if (strongSelf) {
+                                                    [strongSelf->_parameterString appendCharacter:c];
+                                                }
                                             }];
     // Private char; parameter string must follow.
     [dcsEntryState addStateTransitionForCharacterRange:MakeCharacterRange('<', '?')
                                                     to:dcsParamState
                                             withAction:^(unsigned char c) {
-                                                [_privateMarkers appendCharacter:c];
+                                                __strong __typeof(self) strongSelf = weakSelf;
+                                                if (strongSelf) {
+                                                    [strongSelf->_privateMarkers appendCharacter:c];
+                                                }
                                             }];
     // Colon is not allowed here.
     [dcsEntryState addStateTransitionForCharacter:':'
@@ -238,7 +270,10 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     [dcsIntermediateState addStateTransitionForCharacterRange:MakeCharacterRange(' ', '/')
                                                            to:dcsIntermediateState
                                                    withAction:^(unsigned char c) {
-                                                       [_intermediateString appendCharacter:c];
+                                                       __strong __typeof(self) strongSelf = weakSelf;
+                                                       if (strongSelf) {
+                                                           [strongSelf->_intermediateString appendCharacter:c];
+                                                       }
                                                    }];
     // Illegal characters; swallow them up til we get a terminator in the ignore state.
     [dcsIntermediateState addStateTransitionForCharacterRange:MakeCharacterRange('0', '?')
@@ -262,19 +297,28 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     [dcsParamState addStateTransitionForCharacterRange:MakeCharacterRange(' ', '/')
                                                     to:dcsIntermediateState
                                             withAction:^(unsigned char c) {
-                                                [_intermediateString appendCharacter:c];
+                                                __strong __typeof(self) strongSelf = weakSelf;
+                                                if (strongSelf) {
+                                                    [strongSelf->_intermediateString appendCharacter:c];
+                                                }
                                             }];
     // Got a number, save it as a parameter.
     [dcsParamState addStateTransitionForCharacterRange:MakeCharacterRange('0', '9')
                                                     to:dcsParamState
                                             withAction:^(unsigned char c) {
-                                                [_parameterString appendCharacter:c];
+                                                __strong __typeof(self) strongSelf = weakSelf;
+                                                if (strongSelf) {
+                                                    [strongSelf->_parameterString appendCharacter:c];
+                                                }
                                             }];
     // Got a semicolon, save it as a parameter.
     [dcsParamState addStateTransitionForCharacter:';'
                                                to:dcsParamState
                                        withAction:^(unsigned char c) {
-                                           [_parameterString appendCharacter:c];
+                                           __strong __typeof(self) strongSelf = weakSelf;
+                                           if (strongSelf) {
+                                               [strongSelf->_parameterString appendCharacter:c];
+                                           }
                                        }];
     // Initial passthrough character. Entry action will save it.
     [dcsParamState addStateTransitionForCharacterRange:MakeCharacterRange('@', '~')
@@ -284,12 +328,24 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     ////////////////////////////////////////////////////////////////////////////////////////////////
     // Add transitions from ignore state.
     // Only way out of ignore is ST.
-    dcsIgnoreState.entryAction = ^(unsigned char c) { _malformed = YES; };
+    dcsIgnoreState.entryAction = ^(unsigned char c) {
+        __strong __typeof(self) strongSelf = weakSelf;
+        if (strongSelf) {
+            strongSelf->_malformed = YES;
+        }
+    };
     [dcsIgnoreState addStateTransitionForCharacter:VT100CC_ESC
                                                 to:dcsEscapeState
                                         withAction:^(unsigned char character) {
-                                            [self unhook];
+                                            [weakSelf unhook];
                                         }];
+    if (support8BitControlCharacters) {
+        [dcsIgnoreState addStateTransitionForCharacter:VT100CC_C1_ST
+                                                    to:groundState
+                                            withAction:^(unsigned char c) {
+                                                [weakSelf execute];
+                                            }];
+    }
 
     ////////////////////////////////////////////////////////////////////////////////////////////////
     // Add transitions from dcs escape state.
@@ -305,7 +361,7 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     [dcsEscapeState addStateTransitionForCharacter:'\\'
                                                 to:groundState
                                         withAction:^(unsigned char c) {
-                                            [self execute];
+                                            [weakSelf execute];
                                         }];
 
     ////////////////////////////////////////////////////////////////////////////////////////////////
@@ -315,42 +371,51 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     // that cause us to re-enter passthrough after having already been in it so don't do anything
     // destructive here.
     dcsPassthroughState.entryAction = ^(unsigned char c) {
-        [_data appendCharacter:c];
-        if (!_hook) {
-            [self hook];
+        __strong __typeof(self) strongSelf = weakSelf;
+        if (strongSelf) {
+            [strongSelf->_data appendCharacter:c];
+            if (!strongSelf->_hook) {
+                [strongSelf hook];
+            }
         }
     };
     [dcsPassthroughState addStateTransitionForCharacterRange:MakeCharacterRange(VT100CC_NULL,
                                                                                 VT100CC_ETB)
                                                           to:dcsPassthroughState
                                                   withAction:^(unsigned char c) {
-                                                      [self put:c];
+                                                      [weakSelf put:c];
                                                   }];
     [dcsPassthroughState addStateTransitionForCharacter:VT100CC_EM
                                                      to:dcsPassthroughState
                                              withAction:^(unsigned char c) {
-                                                 [self put:c];
+                                                 [weakSelf put:c];
                                              }];
     [dcsPassthroughState addStateTransitionForCharacterRange:MakeCharacterRange(VT100CC_FS,
                                                                                 VT100CC_US)
                                                           to:dcsPassthroughState
                                                   withAction:^(unsigned char c) {
-                                                      [self put:c];
+                                                      [weakSelf put:c];
                                                   }];
     [dcsPassthroughState addStateTransitionForCharacterRange:MakeCharacterRange(' ', '~')
                                                           to:dcsPassthroughState
                                                   withAction:^(unsigned char c) {
-                                                      [self put:c];
+                                                      [weakSelf put:c];
                                                   }];
+    if (support8BitControlCharacters) {
+        [dcsPassthroughState addStateTransitionForCharacter:VT100CC_C1_ST
+                                                         to:groundState
+                                                 withAction:^(unsigned char c) {
+                                                     [weakSelf execute];
+                                                 }];
+    }
 
-    _stateMachine = stateMachine;
-    _stateMachine.groundState = groundState;
+    stateMachine.groundState = groundState;
     return stateMachine;
 }
 
 // Retrieve the token from the state machine's user info dictionary.
 - (VT100Token *)token {
-    return _stateMachine.userInfo[kVT100DCSUserInfoToken];
+    return self.stateMachine.userInfo[kVT100DCSUserInfoToken];
 }
 
 // Save a passthrough character. If the tmux hack weren't here, it would direct to the hook if one
@@ -379,10 +444,23 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
         // could end in a BEL.
         [characterSet addCharactersInRange:NSMakeRange(VT100CC_BEL, 1)];
         [characterSet invert];
-        garbageCharacterSet = [characterSet retain];
+        garbageCharacterSet = characterSet;
     });
-    return (_data.length > kMaxDataLength ||
-            [_data rangeOfCharacterFromSet:garbageCharacterSet].location != NSNotFound);
+    NSInteger offset = MAX(0, _checkedCount);
+    const NSInteger limit = _data.length;
+    if (offset > limit) {
+        offset = MAX(0, limit);
+    }
+    _checkedCount = offset + limit;
+    NSInteger maxLength = kMaxDataLength;
+    if ([_data hasPrefix:@"tmux;\e]1337;File="]) {
+        // Allow file downloads to get really big.
+        maxLength = NSIntegerMax;
+    }
+    return (_data.length > maxLength ||
+            [_data rangeOfCharacterFromSet:garbageCharacterSet
+                                   options:0
+                                     range:NSMakeRange(offset, limit - offset)].location != NSNotFound);
 }
 
 - (void)decodeFromContext:(iTermParserContext *)context
@@ -390,31 +468,37 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
                  encoding:(NSStringEncoding)encoding
                savedState:(NSMutableDictionary *)savedState {
     DLog(@"DCS parser running");
+    _support8BitControlCharacters = (encoding == NSASCIIStringEncoding || encoding == NSISOLatin1StringEncoding);
     static NSString *const kOffset = @"offset";
     if (savedState[kOffset]) {
         iTermParserAdvanceMultiple(context, [savedState[kOffset] intValue]);
     }
-    _stateMachine.userInfo = @{ kVT100DCSUserInfoToken: result };
+    self.stateMachine.userInfo = @{ kVT100DCSUserInfoToken: result };
     result->type = VT100_WAIT;
     while (result->type == VT100_WAIT && iTermParserCanAdvance(context)) {
         if (_hook && !_hookFinished) {
             DLog(@"Sending input to hook %@", _hook);
-            _hookFinished = [_hook handleInput:context token:result];
+            _hookFinished = [_hook handleInput:context
+                  support8BitControlCharacters:_support8BitControlCharacters
+                                         token:result];
+            if (_hookFinished) {
+                [self unhook];
+            }
         } else {
-            [_stateMachine handleCharacter:iTermParserConsume(context)];
-            if ([_stateMachine.currentState.identifier isEqual:@(kVT100DCSStatePassthrough)] &&
+            [self.stateMachine handleCharacter:iTermParserConsume(context)];
+            if ([self.stateMachine.currentState.identifier isEqual:@(kVT100DCSStatePassthrough)] &&
                 [self dataLooksLikeBinaryGarbage]) {
                 result->type = VT100_BINARY_GARBAGE;
             }
         }
-        if (_stateMachine.currentState == _stateMachine.groundState) {
+        if (self.stateMachine.currentState == self.stateMachine.groundState) {
             break;
         }
     }
 
-    _stateMachine.userInfo = nil;
+    self.stateMachine.userInfo = nil;
 
-    if (_stateMachine.currentState != _stateMachine.groundState) {
+    if (self.stateMachine.currentState != self.stateMachine.groundState) {
         if (result->type == VT100_WAIT && !_hook) {
             savedState[kOffset] = @(iTermParserNumberOfBytesConsumed(context));
             iTermParserBacktrack(context);
@@ -424,29 +508,50 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     }
 }
 
+// Only add to this if you're doing something fancy. You can usually add your case to -execute
+// instead.
 - (void)hook {
-    if ([self compactSequence] == MAKE_COMPACT_SEQUENCE(0, 0, 'p') &&
-        [[self parameters] isEqual:@[ @"1000" ]]) {
-        VT100Token *token = _stateMachine.userInfo[kVT100DCSUserInfoToken];
-        if (token) {
-            token->type = DCS_TMUX_HOOK;
-        }
+    switch ([self compactSequence]) {
+    case MAKE_COMPACT_SEQUENCE(0, 0, 'p'):
+        if ([[self parameters] isEqual:@[ @"1000" ]]) {
+            VT100Token *token = self.stateMachine.userInfo[kVT100DCSUserInfoToken];
+            if (token) {
+                token->type = DCS_TMUX_HOOK;
+                _uniqueID = [[[NSUUID UUID] UUIDString] copy];
+                token.string = _uniqueID;
+            }
 
-        [_hook release];
-        _hook = [[VT100TmuxParser alloc] init];
-        _hookFinished = NO;
+            _hook = [[VT100TmuxParser alloc] init];
+            _hookFinished = NO;
+        }
+        break;
+
+        case MAKE_COMPACT_SEQUENCE(0, 0, 'q'): {
+            VT100Token *token = self.stateMachine.userInfo[kVT100DCSUserInfoToken];
+            if (token) {
+                token->type = VT100_SKIP;
+                _uniqueID = [[[NSUUID UUID] UUIDString] copy];
+                token.string = _uniqueID;
+            }
+
+            _hook = [[VT100SixelParser alloc] initWithParameters:[self parameters]];
+            _hookFinished = NO;
+            break;
+        }
     }
 }
 
 - (void)unhook {
-    [_hook autorelease];
     _hook = nil;
+    _uniqueID = nil;
+    [_parameterString deleteCharactersInRange:NSMakeRange(0, _parameterString.length)];
+    self.stateMachine.currentState = self.stateMachine.groundState;
 }
 
 // Force the ground state. Used when force-quitting tmux mode.
 - (void)reset {
     [self unhook];
-    _stateMachine.currentState = _stateMachine.groundState;
+    self.stateMachine.currentState = self.stateMachine.groundState;
 }
 
 // Combines the private-mode character, intermediate character, and first
@@ -497,12 +602,29 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
             }
             break;
 
+        case MAKE_COMPACT_SEQUENCE(0, 0, 'q'):
+            [self unhook];
+            token->type = VT100_SKIP;
+            break;
+
         case MAKE_COMPACT_SEQUENCE(0, 0, 't'):
             if ([_data hasPrefix:@"tmux;"]) {
                 token->type = DCS_TMUX_CODE_WRAP;
                 token.string = [_data substringFromIndex:5];
             }
             break;
+
+        case MAKE_COMPACT_SEQUENCE('=', 0, 's'):
+            if ([_parameterString isEqualToString:@"1"]) {
+                token->type = DCS_BEGIN_SYNCHRONIZED_UPDATE;
+            } else if ([_parameterString isEqualToString:@"2"]) {
+                token->type = DCS_END_SYNCHRONIZED_UPDATE;
+            }
+            break;
+        case MAKE_COMPACT_SEQUENCE(0, '$', 'q'):
+            token->type = DCS_DECRQSS;
+            token.string = [_data substringFromIndex:1];
+            break;
     }
 }
 
@@ -535,16 +657,16 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
     }
 }
 
-- (void)startTmuxRecoveryMode {
+- (void)startTmuxRecoveryModeWithID:(NSString *)dcsID {
     // Put the state machine in the passthrough mode.
     char *fakeControlSequence = "\eP1000p";
     for (int i = 0; fakeControlSequence[i]; i++) {
-        [_stateMachine handleCharacter:fakeControlSequence[i]];
+        [self.stateMachine handleCharacter:fakeControlSequence[i]];
     }
 
     // Replace the hook with one in recovery mode.
-    [_hook release];
     _hook = [[VT100TmuxParser alloc] initInRecoveryMode];
+    _uniqueID = [dcsID copy];
 }
 
 @end
@@ -552,7 +674,7 @@ static NSRange MakeCharacterRange(unsigned char first, unsigned char lastInclusi
 @implementation VT100DCSParser (Testing)
 
 - (VT100DCSState)state {
-    return [(NSNumber *)_stateMachine.currentState.identifier intValue];
+    return [(NSNumber *)self.stateMachine.currentState.identifier intValue];
 }
 
 - (NSString *)data {
